"""Run docker with test commands."""

from __future__ import annotations

from typing import TYPE_CHECKING

import typer

from container_tester.docker_backend import DockerBackend, DockerContainerInfo

if TYPE_CHECKING:
    from container_tester._utils import DockerConfig


def test_container(
    os_name: str,
    name: str,
    command: str = "",
    os_commands: list[str] | None = None,
    *,
    clean: bool = False,
) -> DockerContainerInfo:
    """
    Generate, build, and run a container from provided arguments.

    Args:
        os_name (str): Base OS for the Dockerfile.
        name (str): Identifier for the image and Dockerfile.
        command (str): Command to execute in the container.
        os_commands (list[str]): List of shell commands to include in the
                Dockerfile.
        clean (bool): If True, remove generated artifacts after execution.

    """
    typer.echo(f"{typer.style('Test', fg=typer.colors.GREEN)}: {os_name}")

    docker_test = DockerBackend(
        os_name,
        image_tag=name,
        command=command,
        os_commands=os_commands,
    )

    docker_test.build()
    container = docker_test.run()

    if clean:
        docker_test.remove_image()
        docker_test.remove_container(container.id or "")
        docker_test.remove_dangling()

    return container


def run_config(
    config_list: dict[str, DockerConfig],
    *,
    clean: bool = False,
) -> list[DockerContainerInfo]:
    """
    Generate, build, and run containers from the default config list.

    Args:
        config_list (dict[str, DockerConfig]): Docker image profiles to generate
            files from.
        clean (bool, optional): If True, remove generated files and images
            after execution.

    """
    typer.echo(f"Container Tests: {len(config_list)}")

    info_list = [
        test_container(
            cfg["os_name"],
            tag,
            cfg["command"],
            cfg["os_commands"],
            clean=clean,
        )
        for tag, cfg in config_list.items()
    ]

    return info_list



import re
import shutil
from typing import Annotated

import typer

from container_tester import _utils, app

cli = typer.Typer()
flags = {"json": False, "pretty": False, "clean": False}


@cli.command()
def test_config():
    """Generate, build, and run Docker resources from a docker config."""
    cfg_list = _utils.load_config()
    data = app.run_config(cfg_list, clean=flags["clean"])
    width = shutil.get_terminal_size().columns

    if not data:
        typer.echo("No data.")
        return

    for value in data:
        value.print(json=flags["json"], pretty=flags["pretty"])
        typer.echo("-" * width)


@cli.command()
def test_container(
    os_name: Annotated[
        str,
        typer.Argument(
            help="Base Docker image to initialize from (e.g., 'ubuntu:20.04').",
        ),
    ],
    name: Annotated[
        str,
        typer.Option(help="Custom name for the generated docker image"),
    ] = "",
    command: Annotated[
        str,
        typer.Option(help="Shell command to execute inside the containers."),
    ] = "",
):
    """Generate, build, and run Docker resources from a base image."""
    os_name = os_name.lower().strip()

    if not os_name:
        raise ValueError("The 'os-name' option cannot be empty.")

    if name and not re.fullmatch(r"[a-zA-Z0-9]+", name):
        raise typer.BadParameter(
            f"Invalid name '{name}'. Must contain only letters and digits",
            param_hint="--name",
        )

    data = app.test_container(os_name, name, command, clean=flags["clean"])

    if not data:
        typer.echo("No data.")
        return

    data.print(json=flags["json"], pretty=flags["pretty"])


@cli.callback()
def main(
    *,
    json: Annotated[
        bool,
        typer.Option(help="Show output in json format."),
    ] = False,
    pretty: Annotated[
        bool,
        typer.Option(help="Show output in pretty format."),
    ] = False,
    clean: Annotated[
        bool,
        typer.Option(help="Clean Docker resources after run."),
    ] = False,
) -> None:
    flags["json"] = json
    flags["pretty"] = pretty
    flags["clean"] = clean



from __future__ import annotations

import json
import shutil
import subprocess
import sys
import tomllib as toml
from pathlib import Path
from typing import Any, TypedDict

import typer
from rich.table import Table, box


class DockerConfig(TypedDict):
    """Type a docker config."""

    command: str
    os_name: str
    os_commands: list[str]
    pkg_manager: str


def get_cwd() -> Path | None:
    git_path = shutil.which("git")

    if not git_path:
        return None

    r = subprocess.run(  # noqa: S603
        [git_path, "rev-parse", "--show-toplevel"],
        capture_output=True,
        text=True,
        check=False,
    )
    return Path() if r.returncode else Path(r.stdout.strip())


def load_config() -> dict[str, DockerConfig]:
    file_name = "docker-config.toml"
    user_path = Path(file_name).expanduser()
    default_path = Path(__file__).parent / file_name

    config_path = user_path if user_path.is_file() else default_path

    try:
        with config_path.open("rb") as f:
            return toml.load(f)
    except toml.TOMLDecodeError as e:
        typer.echo(f"Error parsing TOML from {config_path}: {e}", err=True)
        sys.exit(1)
    except OSError as e:
        typer.echo(f"Error reading config file {config_path}: {e}", err=True)
        sys.exit(1)


def format_json(data: Any, *, pretty: bool = False) -> str:
    return json.dumps(data, indent=2 if pretty else None)


def format_table(title: str, data: dict[str, Any], *, pretty: bool = False) -> Table:
    table = Table(
        title=title if pretty else f"{title}:",
        box=box.HEAVY_HEAD if pretty else None,
        title_justify="center" if pretty else "left",
    )
    table.add_column("key", style="cyan" if pretty else None, no_wrap=True)
    table.add_column("value", style="magenta" if pretty else None)

    for key, value in data.items():
        table.add_row(str(key), str(value))

    return table



"""Class to create, build and run docker containers."""

from __future__ import annotations

import re
import sys
import time
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import docker
import rich
import typer
from docker.errors import (
    APIError,
    BuildError,
    ContainerError,
    DockerException,
    ImageNotFound,
    NotFound,
)
from platformdirs import user_cache_dir

from container_tester import _utils

if TYPE_CHECKING:
    from docker import DockerClient
    from docker.models.containers import Container as DockerContainer


@dataclass
class DockerImageInfo:
    """Docker image build data."""

    name: str
    os_name: str
    os_base: str
    os_architecture: str
    size: str

    def print(self, *, json: bool = False, pretty: bool = False) -> None:
        """
        Print image output.

        Args:
            json (bool): Whether to output the data in JSON format.
            pretty (bool): Whether to pretty-print the output.

        """
        data = {
            "name": self.name,
            "os_name": self.os_name,
            "os_base": self.os_base,
            "os_architecture": self.os_architecture,
            "size": self.size,
        }
        if json:
            data_json = _utils.format_json(data)
            rich.print_json(data_json, highlight=pretty)
        elif pretty:
            rich.print(_utils.format_table("image", data, pretty=pretty))
        else:
            typer.echo(data)


@dataclass
class DockerContainerInfo:
    """Docker container data."""

    id: str | None
    name: str
    command: str
    stdout: str
    stderr: str

    def print(self, *, json: bool = False, pretty: bool = False) -> None:
        """
        Print container output.

        Args:
            json (bool): Whether to output the data in JSON format.
            pretty (bool): Whether to pretty-print the output.

        """
        data = {
            "id": self.id,
            "name": self.name,
            "command": self.command,
            "stdout": self.stdout,
            "stderr": self.stderr,
        }
        if json:
            data_json = _utils.format_json(data)
            rich.print_json(data_json, highlight=pretty)
        elif pretty:
            rich.print(_utils.format_table("container", data, pretty=pretty))
        else:
            if self.stdout:
                typer.echo(self.stdout)

            if self.stderr:
                typer.echo(self.stderr, err=True)


class DockerBackend:
    """Manages Dockerfile creation, image building, and container execution."""

    client: DockerClient

    def _get_tag_name(self, name: str) -> str:
        return re.sub(r"[^a-zA-Z0-9]", "_", name)

    def _docker_client(self) -> DockerClient:
        """Return a ready Docker client."""
        try:
            return docker.from_env()
        except DockerException:
            typer.echo(
                "Docker is not running. Please start the Docker daemon and try again.",
                err=True,
            )
            sys.exit(1)

    def __init__(
        self,
        os_name: str,
        *,
        image_tag: str = "",
        command: str = "",
        os_commands: list[str] | None = None,
    ) -> None:
        """Initialize the Docker backend client."""
        self.client = self._docker_client()
        self.os_name = self._get_os_name(os_name)
        self.image_tag = image_tag or self._get_tag_name(self.os_name)
        self.command = command or 'echo "Container is running"'
        self.os_commands = os_commands or []

    def _get_os_name(self, os_name: str) -> str:
        try:
            image = self.client.images.pull(os_name)
            verified_name = image.attrs.get("RepoTags", "")[0]
        except (APIError, ImageNotFound) as e:
            typer.secho(
                f"Failed to retrieve image '{os_name}'.\n{e}",
                fg=typer.colors.RED,
                err=True,
            )
            sys.exit(1)
        else:
            return verified_name

    def _get_template(self) -> str:
        uv_copy = "COPY --from=ghcr.io/astral-sh/uv:latest /uv /uvx /bin/"
        cmds = (
            "\n".join(f"RUN {cmd}" for cmd in self.os_commands)
            if self.os_commands
            else ""
        )

        return f"FROM {self.os_name}\n{uv_copy}\nWORKDIR /app\nCOPY . /app\n{cmds}"

    def _generate_file(self, content: str) -> Path:
        temp_dir = Path(user_cache_dir("container_tester"))
        temp_dir.mkdir(parents=True, exist_ok=True)

        file = temp_dir / "Dockerfile"
        file.write_text(content, encoding="utf-8")

        return file

    def build(self) -> DockerImageInfo:
        """
        Build docker image and optionally remove a tagged Docker image.

        Returns:
            DockerImageInfo: Information about the built Docker image,
            including its name, operating system, architecture, and size.

        Raises:
            SystemExit: If the Docker build fails due to an BuildError,
                APIError, or TypeError.

        """
        content = self._get_template()
        dockerfile = self._generate_file(content)

        try:
            image, _ = self.client.images.build(
                path=str(_utils.get_cwd()),
                dockerfile=str(dockerfile),
                tag=self.image_tag,
                rm=True,  # Necessary to remove intermediate containers.
                forcerm=True,  # Necessary to remove intermediate containers.
            )
            size = image.attrs.get("Size", "") / (1024 * 1024)

            return DockerImageInfo(
                name=self.image_tag,
                os_name=self.os_name,
                os_base=image.attrs["Os"],
                os_architecture=image.attrs["Architecture"],
                size=f"{size:.2f} MB",
            )
        except (BuildError, APIError, TypeError) as e:
            error_msg = getattr(e, "msg", str(e))
            typer.secho(
                f"Failed to build Docker image '{self.image_tag}'.\n{error_msg}",
                fg=typer.colors.RED,
                err=True,
            )
            sys.exit(1)

    def remove_container(self, container_id: str) -> None:
        """
        Remove a Docker container by container_id.

        Args:
            container_id (str): Container name or ID to remove.

        """
        try:
            containers: list[DockerContainer] = self.client.containers.list(all=True)

            for container in containers:
                if container_id in (container.name, container.id):
                    container.stop()
                    container.remove(force=True)
                    break
        except (NotFound, APIError) as e:
            typer.secho(
                f"Failed to remove container '{container_id}'.\n{e}",
                fg=typer.colors.RED,
                err=True,
            )

    def run(self) -> DockerContainerInfo:
        """
        Run a container from a Docker image with the given command.

        Returns:
            DockerContainerInfo: Information about the executed container.
                including its id, name, command, stdout, and stderr.

        Raises:
            SystemExit: If the Docker build fails due to an ContainerError,
                ImageNotFound, or APIError.

        """
        timestamp = int(time.time())
        container_name = f"container_test_{self.image_tag}_{timestamp}"

        try:
            container = self.client.containers.run(
                self.image_tag,
                command=self.command,
                name=container_name,
                detach=True,
                stdout=True,
                stderr=True,
            )
            result = container.wait()

            exit_code = result.get("StatusCode")
            stdout_logs = container.logs(stdout=True, stderr=False).decode()
            stderr_logs = container.logs(stdout=False, stderr=True).decode()
            config = container.attrs.get("Config", {})

            return DockerContainerInfo(
                id=container.id,
                name=container_name,
                command=config.get("Cmd"),
                stdout=stdout_logs.strip(),
                stderr=stderr_logs.strip() if exit_code else "",
            )

        except (ContainerError, ImageNotFound, APIError) as e:
            typer.secho(
                f"Failed to run container from image '{self.image_tag}'.\n{e}",
                fg=typer.colors.RED,
                err=True,
            )
            sys.exit(1)

    def remove_image(self) -> None:
        """
        Remove a Docker image by image-tag.

        Raises:
            DockerException: If the image removal fails due to an API error or
                other Docker-related exception.

        """
        try:
            self.client.images.remove(image=self.image_tag, force=True)
        except (APIError, DockerException, ImageNotFound) as e:
            raise DockerException(f"Failed to remove Docker image.\n{e}") from e

    def remove_dangling(self) -> None:
        """Remove dangling Docker images to free up space."""
        try:
            self.client.images.prune(filters={"dangling": True})
        except DockerException:
            typer.secho(
                "Failed to remove dangling.",
                fg=typer.colors.YELLOW,
                err=True,
            )
